unit Unit18;

(*
version: Embarcadero Delphi XE3 Version 17.0.4625.53395

Posting a record with a null value for a required field fires a 'database error' exception.
Before this exception is being fired, a procedure TField.FocusControl; is executed.
The intention behind that is to set focus on the UI control linked with this TField,
highlighting it for further correction ((ie or cancel the edit, or post a non null value)

The purpose of this demo is to show that LiveBindings does
not support the deFocusControl data event
(ie procedure FocusControl(Field: TFieldRef); is absent in
Data.Bind.DBScope TDataLink desendants), probably
because TBindLinkDataLink is control-agnostic.
To overcome this, traditional DB-aware architecture comes
to rescue: simply create a TFieldDataLink that will hold the link between
LiveBinded DataSource and the UI control, like
    with TFieldDataLink.Create do
    begin
      Control := aEdit;
      FieldName := 'aField';
      DataSource := BindSourceDB1.DataSource;
    end;

 Author: Didier Cabal
         'Bellefont'
         31800 Saint-Gaudens, France
         tl: +33 (0)5 61 89 72 92
         e-mail: didier.cabale+delphi@gmail.com
         web-page: http://didier.cabale.free.fr/delphi.htm

*)

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.DBCtrls, Data.DB,
  Vcl.StdCtrls, Vcl.Mask, Datasnap.DBClient, System.Rtti,
  System.Bindings.Outputs, Vcl.Bind.Editors, Data.Bind.EngExt,
  Vcl.Bind.DBEngExt, Vcl.Buttons, Vcl.Bind.Navigator, Data.Bind.Components,
  Data.Bind.DBScope;

type
  TForm1 = class(TForm)
    ClientDataSet1: TClientDataSet;
    DataSource1: TDataSource;
    ClientDataSet1field1: TStringField;
    GroupBox1: TGroupBox;
    DBEdit1: TDBEdit;
    Label1: TLabel;
    Edit1: TEdit;
    DBNavigator1: TDBNavigator;
    GroupBox2: TGroupBox;
    ClientDataSet2: TClientDataSet;
    BindSourceDB1: TBindSourceDB;
    ClientDataSet2field1: TStringField;
    BindingsList1: TBindingsList;
    NavigatorBindSourceDB1: TBindNavigator;
    Label3: TLabel;
    Label4: TLabel;
    Label2: TLabel;
    Button3: TButton;
    Button4: TButton;
    Button5: TButton;
    Button6: TButton;
    CheckBox1: TCheckBox;
    CheckBox2: TCheckBox;
    Edit2: TEdit;
    Edit3: TEdit;
    Memo1: TMemo;
    LinkControlToField1: TLinkControlToField;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Button6Click(Sender: TObject);
    procedure CheckBox1Click(Sender: TObject);
    procedure CheckBox2Click(Sender: TObject);
    procedure ClientDataSet1PostError(DataSet: TDataSet; E: EDatabaseError;
      var Action: TDataAction);
    procedure ClientDataSet2PostError(DataSet: TDataSet; E: EDatabaseError;
      var Action: TDataAction);
  private
    { Private declarations }
    aFieldDataLink: TFieldDataLink;
  public
    { Public declarations }
  end;

  TMyFieldDataLink = class(TFieldDataLink)
  private
    procedure FocusControl(aField: TFieldRef); override;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  ComCtrls
  ;

procedure TForm1.FormCreate(Sender: TObject);
begin
  ClientDataSet1.CreateDataSet;
  ClientDataSet1.Open;
  ClientDataSet2.CreateDataSet;
  ClientDataSet2.Open;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  ClientDataSet1.Close;
  ClientDataSet2.Close;
//  aFieldDataLink.Free;
  FreeAndNil(aFieldDataLink);
end;

procedure TForm1.Button3Click(Sender: TObject);
begin
  // 1. focus outside DBEdit1
  Edit1.SetFocus;
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
  ClientDataSet1.Append;
  ClientDataSet1.Post;
end;

procedure TForm1.Button5Click(Sender: TObject);
begin
  // 1. focus outside Edit3
  Edit2.SetFocus;
end;

procedure TForm1.Button6Click(Sender: TObject);
begin
//  ClientDataSet2.Edit;
  ClientDataSet2.Append;
  ClientDataSet2.Post;
end;

procedure TForm1.CheckBox1Click(Sender: TObject);
begin
  // 2. create a TFieldDataLink that will eventually execute
  // the procedure TFieldDataLink.FocusControl(Field: TFieldRef);
  Button5Click(nil);
  if TCheckBox(Sender).Checked then
  begin
    if Assigned(aFieldDataLink) then
      exit;
    aFieldDataLink := TFieldDataLink.Create;
    with aFieldDataLink do
    begin
      Control := Edit3;
      FieldName := 'field1';
      DataSource := BindSourceDB1.DataSource;
    end;
  end
  else
    FreeAndNil(aFieldDataLink);
end;

procedure TForm1.ClientDataSet1PostError(DataSet: TDataSet; E: EDatabaseError;
  var Action: TDataAction);
begin
   ShowMessage('note field1 editor gets the focus -> happy !');
end;

procedure TForm1.ClientDataSet2PostError(DataSet: TDataSet; E: EDatabaseError;
  var Action: TDataAction);
begin
   ShowMessage('note field1 editor focus depends on TFieldDataLink existence');
end;

procedure TForm1.CheckBox2Click(Sender: TObject);
begin
  // 2. create a TFieldDataLink that will eventually execute the procedure TFieldDataLink.FocusControl(Field: TFieldRef);
  if TCheckBox(Sender).Checked then
  begin
    if Assigned(aFieldDataLink) then
      exit;
    aFieldDataLink := TMyFieldDataLink.Create;
    with aFieldDataLink do
    begin
      Control := Edit3;
      FieldName := 'field1';
      DataSource := BindSourceDB1.DataSource;
    end;
  end
  else
    FreeAndNil(aFieldDataLink);
end;

{ TMyFieldDataLink }

procedure TMyFieldDataLink.FocusControl(aField: TFieldRef);

  procedure SetFocus(aControl: TWinControl);
  begin
    aField^ := nil;
    aControl.SetFocus;
  end;

var
   ParentControl: TWinControl;
begin
  if (aField^ <> nil) and (aField^ = Field) and (Control is TWinControl) then
    if TWinControl(Control).CanFocus then
      SetFocus(TWinControl(Control))
    //////////////////////////////////////////
    ///   added by DCA on 10/03/2012
    else
    begin
      ParentControl := TWinControl(Control).Parent;
      if Assigned(ParentControl) and (ParentControl is TTabSheet) then
      begin
         TTabSheet(ParentControl).PageControl.ActivePage := TTabSheet(ParentControl);
         if TWinControl(Control).CanFocus then
            SetFocus(TWinControl(Control));
      end;
    end;
    ///  end of adding
end;

end.
